#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

#define WARP_SIZE 32

__global__ void optimized_concat_kernel(
    const float* __restrict__ x,
    const float* __restrict__ hidden,
    float* __restrict__ combined,
    const int batch_size,
    const int x_size,
    const int hidden_size
) {
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int stride = blockDim.x * gridDim.x;
    const int total_width = x_size + hidden_size;
    
    for (int idx = tid; idx < batch_size * total_width; idx += stride) {
        const int row = idx / total_width;
        const int col = idx % total_width;
        if (row < batch_size) {
            if (col < x_size) {
                combined[row * total_width + col] = x[row * x_size + col];
            } else {
                int col_hidden = col - x_size;
                combined[row * total_width + col] = hidden[row * hidden_size + col_hidden];
            }
        }
    }
}

torch::Tensor module_fn_cuda(
    torch::Tensor x,
    torch::Tensor i2h_weight,
    torch::Tensor i2h_bias,
    torch::Tensor h2o_weight,
    torch::Tensor h2o_bias,
    torch::Tensor hidden
) {
    x = x.contiguous();
    i2h_weight = i2h_weight.contiguous();
    i2h_bias = i2h_bias.contiguous();
    h2o_weight = h2o_weight.contiguous();
    h2o_bias = h2o_bias.contiguous();
    hidden = hidden.contiguous();

    const int batch_size = x.size(0);
    const int x_size = x.size(1);
    const int hidden_size = hidden.size(1);

    auto options = torch::TensorOptions().dtype(x.dtype()).device(x.device());
    auto combined = torch::empty({batch_size, x_size + hidden_size}, options);

    // Optimize grid and block dimensions
    const int threads = 512;
    const int blocks = std::min(65535, (batch_size * (x_size + hidden_size) + threads - 1) / threads);

    optimized_concat_kernel<<<blocks, threads>>>(
        x.data_ptr<float>(),
        hidden.data_ptr<float>(),
        combined.data_ptr<float>(),
        batch_size,
        x_size,
        hidden_size
    );

    // Compute hidden = tanh(linear(combined, i2h_weight, i2h_bias))
    auto hidden_new = torch::tanh(torch::addmm(i2h_bias, combined, i2h_weight.t()));

    // Compute output = linear(hidden_new, h2o_weight, h2o_bias)
    auto output = torch::addmm(h2o_bias, hidden_new, h2o_weight.t());

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &module_fn_cuda, "Module forward (CUDA)");
}